/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2016 by 3rd Designing Center
 *              Imageing System Development Division RICOH COMPANY, LTD.
 *  Copyright (C) 2017-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: target_inithook.c 698 2019-07-28 13:58:54Z roi $
 */

#include <sil.h>
#include "stm32l4xx.h"

/*
 *  ץstm32l4r5-nucleo144ѡ
 */

#define sil_modw_mem(addr, mask, val)	sil_wrw_mem((addr), ((sil_rew_mem(addr) & ~(mask)) | (val)))
#define sil_andw_mem(addr, mask)		sil_wrw_mem((addr), (sil_rew_mem(addr) & ~(mask)))
#define sil_orw_mem(addr, val)			sil_wrw_mem((addr), (sil_rew_mem(addr) | (val)))


#define HSI_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define PLL_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define MSI_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define LSE_TIMEOUT_VALUE          (5000*1000)	/* 5 s    */
#define HSI48_TIMEOUT_VALUE        (21000)		/* 2 ms (minimum Tick + 1) */
#define DBP_TIMEOUT_VALUE          (100*1000)	/* 100 ms */
#define PWR_FLAG_TIMEOUT_VALUE     (50)			/* 50micro sec */
#define CLOCKSWITCH_TIMEOUT_VALUE  (5000000U)	/* 5 s    */

#ifdef USBCLOCK_HSI48
#define RCC_MSIRANGE               RCC_CR_MSIRANGE_6
#else
#define RCC_MSIRANGE               RCC_CR_MSIRANGE_11
#endif
#define MAX_MSIRANGE               12

#if !defined(MSI_RANGE)
#define MSI_RANGE                  0
#endif

#if !defined(HSE_VALUE)
#define HSE_VALUE                  8000000  /* Value of the External oscillator in Hz */
#endif /* HSE_VALUE */
#if !defined(HSI_VALUE)
#define HSI_VALUE                  16000000	/* Value of the Internal oscillator in Hz*/
#endif /* HSI_VALUE */

/* PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N */
#ifdef USBCLOCK_HSI48
#define PLL_M      1
#else
#define PLL_M      12
#endif
#define PLL_N      60

/* SYSCLK = PLL_VCO / PLL_P */
#define PLL_P      7

/* USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ */
#define PLL_Q      2

/* I2SCLK =  PLLVCO / PLLR */                          
#define PLL_R      2	/* Possible value between 2 and 7 */

#define RCC_MSICALIBRATION_DEFAULT     0		/* Default MSI calibration trimming value */


/*
 *  VOLTAGE SCALE
 */
#define PWR_FLAG_SETTING_DELAY_US      50				/* Time out value for REGLPF and VOSF flags setting */
#define PWR_REGULATOR_VOLTAGE_SCALE1_BOOST  0x00000000	/* Voltage scaling range 1 boost mode  */
#define PWR_REGULATOR_VOLTAGE_SCALE1   PWR_CR1_VOS_0    /* Voltage scaling range 1 */
#define PWR_REGULATOR_VOLTAGE_SCALE2   PWR_CR1_VOS_1    /* Voltage scaling range 2 */

/*
 *  defgroup RCCEx_CRS_SynchroSource RCCEx CRS SynchroSource
 */
#define RCC_CRS_SYNC_SOURCE_GPIO       0x00000000U             /*!< Synchro Signal source GPIO */
#define RCC_CRS_SYNC_SOURCE_LSE        CRS_CFGR_SYNCSRC_0      /*!< Synchro Signal source LSE */
#define RCC_CRS_SYNC_SOURCE_USB        CRS_CFGR_SYNCSRC_1      /*!< Synchro Signal source USB SOF (default)*/

/*
 *  defgroup RCCEx_USB_Clock_Source USB Clock Source
 */
#define RCC_USBCLKSOURCE_HSI48         0x00000000
#define RCC_USBCLKSOURCE_PLLSAI1       RCC_CCIPR_CLK48SEL_0
#define RCC_USBCLKSOURCE_PLL           RCC_CCIPR_CLK48SEL_1
#define RCC_USBCLKSOURCE_MSI           RCC_CCIPR_CLK48SEL

#ifdef USBCLOCK_HSI48
#define RCC_USBCLKSOURCE               RCC_USBCLKSOURCE_HSI48
#else
#define RCC_USBCLKSOURCE               RCC_USBCLKSOURCE_MSI
#endif

/*
 *  defgroup RCCEx_CRS_SynchroDivider RCCEx CRS SynchroDivider
 */
#define RCC_CRS_SYNC_DIV1              0x00000000			/* Synchro Signal not divided (default) */
#define RCC_CRS_SYNC_DIV2              CRS_CFGR_SYNCDIV_0	/* Synchro Signal divided by 2 */
#define RCC_CRS_SYNC_DIV4              CRS_CFGR_SYNCDIV_1	/* Synchro Signal divided by 4 */
#define RCC_CRS_SYNC_DIV8              (CRS_CFGR_SYNCDIV_1 | CRS_CFGR_SYNCDIV_0)	/* Synchro Signal divided by 8 */
#define RCC_CRS_SYNC_DIV16             CRS_CFGR_SYNCDIV_2	/* Synchro Signal divided by 16 */
#define RCC_CRS_SYNC_DIV32             (CRS_CFGR_SYNCDIV_2 | CRS_CFGR_SYNCDIV_0)	/* Synchro Signal divided by 32 */
#define RCC_CRS_SYNC_DIV64             (CRS_CFGR_SYNCDIV_2 | CRS_CFGR_SYNCDIV_1)	/* Synchro Signal divided by 64 */
#define RCC_CRS_SYNC_DIV128            CRS_CFGR_SYNCDIV		/* Synchro Signal divided by 128 */

/*
 *  defgroup RCCEx_CRS_ErrorLimitDefault RCCEx CRS ErrorLimitDefault
 */
#define RCC_CRS_ERRORLIMIT_DEFAULT     0x00000022U   /*!< Default Frequency error limit */

/*
 *  MSIRANGEбåơ֥
 */
static const uint32_t MSIRangeTable[MAX_MSIRANGE] = {100000, 200000, 400000, 800000, 1000000, 2000000, \
                                      4000000, 8000000, 16000000, 24000000, 32000000, 48000000};

uint32_t SystemFrequency;


/*
 *  VOLTAGE SCALE
 *  param    VoltageScaling
 *           This parameter can be one of the following values:
 *            @arg @ref PWR_REGULATOR_VOLTAGE_SCALE1 Regulator voltage output range 1 mode,
 *                                                typical output voltage at 1.2 V,
 *                                                system frequency up to 80 MHz.
 *            @arg @ref PWR_REGULATOR_VOLTAGE_SCALE2 Regulator voltage output range 2 mode,
 *                                                typical output voltage at 1.0 V,
 *                                                system frequency up to 26 MHz.
 *  return  ｪλʤE_OK֤
 */
static ER
PWRE_ControlVoltageScaling(uint32_t VoltageScaling)
{
	uint32_t tick = 0;

	if(VoltageScaling != PWR_REGULATOR_VOLTAGE_SCALE1
			&& VoltageScaling != PWR_REGULATOR_VOLTAGE_SCALE2 && VoltageScaling != PWR_REGULATOR_VOLTAGE_SCALE1_BOOST)
		return E_PAR;

	if(VoltageScaling == PWR_REGULATOR_VOLTAGE_SCALE1_BOOST){
		/* If current range is range 2 */
		if((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_VOS) == PWR_REGULATOR_VOLTAGE_SCALE2){
			/* Make sure Range 1 Boost is enabled */
			sil_andw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);

			/* Set Range 1 */
			sil_modw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_VOS, PWR_REGULATOR_VOLTAGE_SCALE1);

			/* Wait until VOSF is cleared */
			tick = (PWR_FLAG_SETTING_DELAY_US * (4000000 / 1000000));
			while((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_SR2)) & PWR_SR2_VOSF) != 0){
				if(--tick == 0)
					return E_TMOUT;
			}
		}
		/* If current range is range 1 normal or boost mode */
		else{
			/* Enable Range 1 Boost (no issue if bit already reset) */
			sil_andw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);
		}
	}
	else if(VoltageScaling == PWR_REGULATOR_VOLTAGE_SCALE1){
		/* If current range is range 2 */
		if((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_VOS) == PWR_REGULATOR_VOLTAGE_SCALE2){
			/* Make sure Range 1 Boost is disabled */
			sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);
			/* Set Range 1 */
			sil_modw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_VOS, PWR_REGULATOR_VOLTAGE_SCALE1);

			/* Wait until VOSF is cleared */
			tick = (PWR_FLAG_SETTING_DELAY_US * (4000000 / 1000000));
			while((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_SR2)) & PWR_SR2_VOSF) != 0){
				if(--tick == 0)
					return E_TMOUT;
			}
		}
		/* If current range is range 1 normal or boost mode */
		else{
			/* Disable Range 1 Boost (no issue if bit already set) */
			sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);
		}
	}
	else{
		/* Set Range 2 */
		sil_modw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_VOS, PWR_REGULATOR_VOLTAGE_SCALE2);
		/* No need to wait for VOSF to be cleared for this transition */
		/* PWR_CR5_R1MODE bit setting has no effect in Range 2        */
	}
	return E_OK;
}

/*
 *  MSIRANGE˹碌ŬڤFLASH LATENCYꤹ
 *  param   msirange MSIRANG
 *  return  ｪλʤE_OK֤
 */
static ER
RCC_SetFlashLatencyFromMSIRange(unsigned int msirange)
{
	uint32_t vos = 0;
	uint32_t latency = FLASH_ACR_LATENCY_0WS;	/* default value 0WS */
	uint32_t sysclk  = MSIRangeTable[(msirange >> 4)];

	if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1)) & RCC_APB1ENR1_PWREN) != 0){
		vos = sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_VOS;
	}
	else{
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);
		vos = sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_VOS;
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);
	}

	if(vos == PWR_CR1_VOS_0){		/* VOLTAGE_SCALE1 */
	    if(sysclk > 16000000){		/* 16MHz */
			if(sysclk > 32000000)	/* 32MHz */
				latency = FLASH_ACR_LATENCY_2WS; /* 2WS */
			else					/* 24MHzޤ 32MHz */
				latency = FLASH_ACR_LATENCY_1WS; /* 1WS */
		}
		/* ¾ 0WS */
	}
	else{							/* VOLTAGE_SCALE2 */
		if(sysclk >= 16000000)		/* 16MHzʾ */
			latency = FLASH_ACR_LATENCY_2WS; /* 2WS */
		else{
			if(sysclk == 8000000)	/* 8MHz */
				latency = FLASH_ACR_LATENCY_1WS; /* 1WS */
			/* ¾ 0WS */
		}
	}
	/*
	 *  FLASH LATENCY
	 */
	sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, latency);

	/*
	 *  FLASH LATENCYǧ
	 */
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != latency)
		return E_SYS;
	return E_OK;
}

/*
 *  祷ƥ९å
 *  ͤϰʲ̤
 *      System Clock source            = PLL (MSI)
 *      SYSCLK(Hz)                     = 120000000
 *      HCLK(Hz)                       = 120000000
 *      AHB Prescaler                  = 1
 *      APB1 Prescaler                 = 1
 *      APB2 Prescaler                 = 1
 *      MSI Frequency(Hz)              = 48000000(4000000)
 *      PLL_M                          = 12(1)
 *      PLL_N                          = 60
 *      PLL_R                          = 2
 *      PLL_P                          = 7
 *      PLL_Q                          = 2
 *      Flash Latency(WS)              = 5
 *  return  ｪλʤE_OK֤
 */
static ER
SystemClock_Config(void)
{
	ER       ercd;
	uint32_t msi_range = 0;
	uint32_t tick = 0;
	uint32_t pwrclkchanged = 0;
#ifdef USBCLOCK_HSI48
	uint32_t value = 0;
#endif

	if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1)) & RCC_APB1ENR1_PWREN) == 0){
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);
		pwrclkchanged = 1;
	}
	if((ercd = PWRE_ControlVoltageScaling(PWR_REGULATOR_VOLTAGE_SCALE1_BOOST)) != E_OK)
		return ercd;

	/*
	 *  LSE ENABLE
	 */
	if((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_DBP) == 0){
		sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_DBP);
		while((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_DBP) == 0){
			if(tick > DBP_TIMEOUT_VALUE)
				return E_TMOUT;
			tick++;
			sil_dly_nse(1000);
		}
	}

	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSEON);
	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & RCC_BDCR_LSERDY) == 0){
		if(tick > LSE_TIMEOUT_VALUE)
			return E_TMOUT;
		tick++;
		sil_dly_nse(1000);
    }

	if(pwrclkchanged)
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);

	/*
	 * Disable the LSE Clock Security System.
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSECSSON);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CIER), RCC_CIFR_LSECSSF);

#ifdef USBCLOCK_HSI48
      /* Enable the Internal Low Speed oscillator (HSI48). */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CRRCR), RCC_CRRCR_HSI48ON);
	tick = 0;

	/* Wait till HSI48 is ready */
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CRRCR)) & RCC_CRRCR_HSI48RDY) == 0){
		if(tick > (HSI48_TIMEOUT_VALUE)){
			return E_TMOUT;
		}
		tick++;
		sil_dly_nse(1000);
	}
#endif

	/*
	 *  MSI
	 */
    if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) == RCC_CFGR_SWS_MSI){
		/*
		 *  MSI򥷥ƥ९åȤƻѻMSIΥ󥸤ФRCC_MSIRANGEꤷʤ
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRGSEL) != 0)
			msi_range = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRANGE;
		else
			msi_range = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR)) & RCC_CSR_MSISRANGE) >> 4;
		if(RCC_MSIRANGE > msi_range){
			/*
			 *  ⤤硢FLASH LATENCY
			 */
			if(RCC_SetFlashLatencyFromMSIRange(RCC_MSIRANGE) != E_OK){
				return E_TMOUT;
			}

			/*
			 *  RCC_MSIRANGE˺
			 */
			sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, RCC_MSIRANGE);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);
		}
		else{
			/*
			 *  㤤硢RCC_MSIRANGE˺
			 */
			sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, RCC_MSIRANGE);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);

			/*
			 *  FLASH LATENCY
			 */
			if(RCC_SetFlashLatencyFromMSIRange(RCC_MSIRANGE) != E_OK){
				return E_TMOUT;
			}
		}
    }
    else{
		/*
		 *  MSISYSCLKǤʤ硢ư
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSION);

        /*
		 *  MSI READYԤ
		 */
        while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRDY) == 0){
			if(tick > MSI_TIMEOUT_VALUE){
				return E_TMOUT;
			}
			tick++;
			sil_dly_nse(1000);
		}

		/*
		 *  RCC_MSIRANGE
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, RCC_MSIRANGE);
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);
	}

	/*
	 *  Enable MSI Auto-calibration through LSE
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIPLLEN);

	/*
	 *  Select MSI output as USB clock source
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CCIPR), RCC_CCIPR_CLK48SEL, RCC_USBCLKSOURCE);

    /*
	 *  PLL
	 */
    if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL){
		/*
		 *  PLLSYSCLKǤʤ硢PLL
		 */
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);

		/*
		 *  PLLԤ
		 */
		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) != 0){
			if(tick > PLL_TIMEOUT_VALUE){
				return E_TMOUT;
			}
			tick++;
			sil_dly_nse(1000);
 		}

		/*
		 *  PLLå
		 */
		sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), ((PLL_M - 1) << 4) | (PLL_N << 8) | ((PLL_P >> 4) << 17) | \
                   RCC_PLLCFGR_PLLSRC_MSI | (((PLL_Q >> 1) - 1) << 21) | (((PLL_R >> 1) - 1) << 25));

		/*
		 *  PLLƵư
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);

		/*
		 *  PLL ƥ९å
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), RCC_PLLCFGR_PLLREN);

		/*
		 *  PLL READYԤ
		 */
        tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) == 0){
			if(tick > PLL_TIMEOUT_VALUE){
				return E_TMOUT;
			}
			tick++;
			sil_dly_nse(1000);
		}
	}
	else
		return E_SYS;

	/*
	 *  ׵FLASH LATENCY⤤硢FLASHLATENCYԤ
	 */
	if(FLASH_ACR_LATENCY_3WS > (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  LATENCYꤷǧ
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLASH_ACR_LATENCY_3WS);
		if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLASH_ACR_LATENCY_3WS){
			return E_SYS;
		}
	}

	/*
	 *  HCLKʬ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_HPRE, RCC_CFGR_HPRE_DIV2);

	/*
	 *  SYSCLKPLLѹ塢ǧ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_SW, RCC_CFGR_SW_PLL);

	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL){
		if(tick > CLOCKSWITCH_TIMEOUT_VALUE){
			return E_TMOUT;
		}
		tick++;
		sil_dly_nse(1000);
	}

	/*
	 *  ׵FLASH LATENCY㤤硢FLASHLATENCYԤ
	 */
	if(FLASH_ACR_LATENCY_3WS < (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  LATENCYꤷǧ
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLASH_ACR_LATENCY_3WS);
		if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLASH_ACR_LATENCY_3WS){
			return E_SYS;
		}
	}
	SystemFrequency = 120000000;

	/*
	 *  PCLK1ʬͤ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE1, RCC_CFGR_PPRE1_DIV1);

	/*
	 *  PCLK2ʬͤ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE2, RCC_CFGR_PPRE2_DIV1);

	/*
	 *  ׵FLASH LATENCY⤤硢FLASHLATENCYԤ
	 */
	if(FLASH_ACR_LATENCY_5WS > (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  LATENCYꤷǧ
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLASH_ACR_LATENCY_5WS);
		if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLASH_ACR_LATENCY_5WS){
			return E_SYS;
		}
	}

	/*
	 *  HCLKʬ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_HPRE, RCC_CFGR_HPRE_DIV1);

	/*
	 *  ׵FLASH LATENCY㤤硢FLASHLATENCYԤ
	 */
	if(FLASH_ACR_LATENCY_5WS < (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  LATENCYꤷǧ
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLASH_ACR_LATENCY_5WS);
		if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLASH_ACR_LATENCY_5WS){
			return E_SYS;
		}
	}

#ifdef USBCLOCK_HSI48
	/*
	 *  CRSͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_CRSEN);

	/*
	 *  CRSꥻåȸ塢ꥻåȲ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_CRSRST);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_CRSRST);

	/* Set the SYNCDIV[2:0] bits according to Prescaler value */
	/* Set the SYNCSRC[1:0] bits according to Source value */
	/* Set the SYNCSPOL bit according to Polarity value */
	value = (RCC_CRS_SYNC_DIV1 | RCC_CRS_SYNC_SOURCE_USB | /*pInit->Polarity*/ 0);
	/* Set the RELOAD[15:0] bits according to ReloadValue value */
	value |= ((48000000 / 1000) - 1);

	/* Set the FELIM[7:0] bits according to ErrorLimitValue value */
	value |= (RCC_CRS_ERRORLIMIT_DEFAULT << 16);
	sil_wrw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CFGR), value);

	/* Adjust HSI48 oscillator smooth trimming */
	/* Set the TRIM[5:0] bits according to RCC_CRS_HSI48CalibrationValue value */
	sil_modw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CR), CRS_CR_TRIM, (0x20 << 8));

	/* START AUTOMATIC SYNCHRONIZATION*/

	/* Enable Automatic trimming & Frequency error counter */
	sil_orw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CR), (CRS_CR_AUTOTRIMEN | CRS_CR_CEN));
#endif

	/*
	 *  USBV򥪥
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);
	sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR2), PWR_CR2_USV);		/* for USB */
	return E_OK;
}

/*
 *  MSIˤ륯å
 *  ͤϰʲ̤
 *      System Clock source            = MSI
 *      SYSCLK(Hz)                     = msirangeΥå
 *      HCLK(Hz)                       = msirangeΥå
 *      AHB Prescaler                  = 1()
 *      APB1 Prescaler                 = 1()
 *      APB2 Prescaler                 = 1()
 *      MSI Frequency(Hz)              = msirangeΥå
 *      PLL_M                          = 1
 *      PLL_N                          = 60
 *      PLL_R                          = 2
 *      PLL_P                          = 7
 *      PLL_Q                          = 2
 *      Flash Latency(WS)              = msirange˽äLATENCY
 *  param   msirange
 *  return  ｪλʤE_OK֤
 */
static ER
SystemLowClock_Config(uint8_t msirange)
{
	uint32_t tick = 0;
	ER ercd = E_OK;

	if(msirange == 0 || msirange >= MAX_MSIRANGE)
		return E_PAR;

	if(MSIRangeTable[msirange] == SystemFrequency)
		return E_OK;

	if((ercd = PWRE_ControlVoltageScaling(PWR_REGULATOR_VOLTAGE_SCALE1)) != E_OK)
		return ercd;

	/*
	 *  MSI
	 */
    if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) == RCC_CFGR_SWS_MSI){
		uint32_t msi_range;
		/*
		 *  MSI򥷥ƥ९åȤƻѻMSIΥ󥸤Фmsirangeꤷʤ
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRGSEL) != 0)
			msi_range = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRANGE;
		else
			msi_range = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR)) & RCC_CSR_MSISRANGE) >> 4;
		if((msirange<<4) > msi_range){
			/*
			 *  ⤤硢FLASH LATENCY
			 */
			if(RCC_SetFlashLatencyFromMSIRange((msirange<<4)) != E_OK){
				return E_TMOUT;
			}

			/*
			 *  msirange˺
			 */
			sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, (msirange<<4));
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);
		}
		else{
			/*
			 *  㤤硢msirange˺
			 */
			sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, (msirange<<4));
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);

			/*
			 *  FLASH LATENCY
			 */
			if(RCC_SetFlashLatencyFromMSIRange((msirange<<4)) != E_OK){
				return E_TMOUT;
			}
		}
    }
    else{
		/*
		 *  MSISYSCLKǤʤ硢ư
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSION);

        /*
		 *  MSI READYԤ
		 */
        while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_MSIRDY) == 0){
			if(tick > MSI_TIMEOUT_VALUE){
				return E_TMOUT;
			}
			tick++;
			sil_dly_nse(1000);
		}

		/*
		 *  msirange
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRGSEL);
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSIRANGE, (msirange<<4));
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_MSITRIM, RCC_MSICALIBRATION_DEFAULT << 8);

		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_SW, RCC_CFGR_SW_MSI);

		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) != RCC_CFGR_SWS_MSI){
			if(tick > CLOCKSWITCH_TIMEOUT_VALUE)
				return E_TMOUT;
			tick++;
			sil_dly_nse(1000);
		}
	}

	if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLON) != 0){
		/*
		 *  PLL
		 */
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);

		/*
		 *  SAIå
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & (RCC_CR_PLLSAI1RDY | RCC_CR_PLLSAI2RDY)) == 0){
			sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), RCC_PLLCFGR_PLLSRC);
		}

		/*
		 *  PLLP/PLLQ/PLLR
		 */
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), (RCC_PLLCFGR_PLLPEN | RCC_PLLCFGR_PLLQEN | RCC_PLLCFGR_PLLREN));

		/*
		 *  PLLԤ
		 */
		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) != 0){
			if(tick > PLL_TIMEOUT_VALUE)
				return E_TMOUT;
			tick++;
			sil_dly_nse(1000);
		}
	}
	SystemFrequency = MSIRangeTable[msirange];

	if(SystemFrequency <= 20000000){
		ercd = PWRE_ControlVoltageScaling(PWR_REGULATOR_VOLTAGE_SCALE2);
	}

	/*
	 *  USBV򥪥
	 */
	sil_andw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR2), PWR_CR2_USV);		/* for USB */
	return ercd;
}

/*
 *  ưåμФ
 */
uint32_t
sysclock_init_value(void)
{
#if MSI_RANGE != 0
	return MSIRangeTable[MSI_RANGE];
#else
	return 120000000;	/* 120 MHz */
#endif
}

/*
 *  ƥ९åѹ
 */
ER
sysclock_change(uint8_t range)
{
	if(range == 0)
		return SystemClock_Config();
	else
		return SystemLowClock_Config(range);
}

/*
 *  ƥ९åե졼
 */
ER
sysclock_config(int mode)
{
	return SystemClock_Config();
}


#ifndef TOPPERS_RAM_EXEC
/*
 *  ٥ΥåȰ¸ν
 *
 *  ȥåץ⥸塼ǡν˸ƤӽФ롥
 */
void hardware_init_hook(void)
{
	/* FPU settings ------------------------------------------------------------*/
	sil_orw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_CPACR), ((3UL << 10*2)|(3UL << 11*2)));	/* set CP10 and CP11 Full Access */

	/*
	 * RCCΥꥻåȤȥե졼
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_MSION);
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), 0x00000000);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR),
			 (RCC_CR_PLLSAI2ON | RCC_CR_PLLSAI1ON | RCC_CR_PLLON | RCC_CR_HSEBYP | RCC_CR_HSEON));
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), RCC_PLLCFGR_PLLN_4);

	/* Reset HSEBYP bit */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_HSEBYP);

	/*
	 *  ߶ػ
	 */
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CIER), 0x00000000);

	if(sysclock_change(MSI_RANGE) != E_OK){
		while(1){}
	}

	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_PWREN);
	sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR2), PWR_CR2_IOSV);	/* for LPUART */

	/*
	 *  PCLK
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CCIPR), ~RCC_CCIPR_CLK48SEL);
}
#endif

